/*
 * Copyright (C) 2013 Google Inc. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Modified by Octoshape 2015 to reduce the responsibility of this class to local playback.
  */

package octoshape.android.mediaplayer;

import android.annotation.SuppressLint;
import android.content.Intent;
import android.content.res.Configuration;
import android.graphics.Point;
import android.media.AudioManager;
import android.media.MediaPlayer;
import android.media.MediaPlayer.OnCompletionListener;
import android.media.MediaPlayer.OnErrorListener;
import android.media.MediaPlayer.OnPreparedListener;
import android.net.Uri;
import android.os.Build;
import android.os.Bundle;
import android.os.Handler;
import android.support.v4.app.NavUtils;
import android.support.v7.app.ActionBarActivity;
import android.support.v7.widget.Toolbar;
import android.text.method.ScrollingMovementMethod;
import android.util.Log;
import android.view.Menu;
import android.view.MenuItem;
import android.view.MotionEvent;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.View.OnTouchListener;
import android.view.WindowManager;
import android.widget.ImageView;
import android.widget.ProgressBar;
import android.widget.RelativeLayout;
import android.widget.SeekBar;
import android.widget.SeekBar.OnSeekBarChangeListener;
import android.widget.TextView;

import com.google.android.gms.cast.MediaInfo;
import com.google.android.gms.cast.MediaMetadata;
import octoshape.android.CastApplication;
import octoshape.android.R;
import com.google.sample.cast.refplayer.settings.CastPreference;
import com.google.sample.cast.refplayer.utils.Utils;

import org.json.JSONException;

import java.io.IOException;
import java.util.HashMap;
import java.util.Timer;
import java.util.TimerTask;

import octoshape.android.stream.SeekType;
import octoshape.android.stream.StreamManager;
import octoshape.android.stream.StreamPlayerWrapper;

/**
 * This class has been modified from the original to handle only local playing.
 * The functionality of switching between local and remote playing has been moved to {@linkplain octoshape.android.mediaplayer.PlayerManager}
 *
 * Compared to the original class, this version uses MediaPlayer, so we can tap into the listeners.
 */
public class LocalPlayerActivity extends ActionBarActivity implements SurfaceHolder.Callback,
        IMediaPlayer {

    private static final String TAG = "lpa_LocalPlayer";
    private static final String MPL_TAG = "MPL";

    // UI Elements
    private SurfaceView surfaceView;
    private SurfaceHolder surfaceHolder;
    MediaPlayer mediaPlayer;

    private TextView titleView;
    private TextView descriptionView;
    private TextView authorView;

    private SeekBar seekBar;
    TextView startText;
    TextView endText;

    private ImageView playPause;

    private ProgressBar loadingProgressBar;
    private View controllersView;
    private View containerView;

    // UI RELATED
    private boolean areControllersVisible;
    private Timer seekbarTimer;
    private Timer controllersTimer;
    private Point displaySize;
    private final float aspectRatio = 72f / 128;
    private final Handler handler = new Handler();

    // MEDIA PLAYER RELATED
    private MediaInfo selectedMediaInfo;
    private int duration;
    private int seekOffset;
    private int latency = 0;
    private IMediaPlayer.PlaybackState playbackState;

    boolean isLive = true;
    private SeekType seekType = SeekType.NO_SEEK;

    /**
     * Class member which tracks the state of the {@linkplain #mediaPlayerState}.
     */
    private PlayerStates mediaPlayerState = PlayerStates.UNINIT;
    private String stream;
    private int initialPosition = 0;
    private boolean notYetPreparedForFirstTime = true;


    /**
     * Enum for the media player states, to shadow the states and make sure we don't call, e.g.,
     * {@linkplain android.media.MediaPlayer#stop()} in an invalid state. We are interested only in the three states:
     * <ul>
     *     <li>{@linkplain #UNINIT} if the media player is uninitialized (e.g., after calling {@linkplain android.media.MediaPlayer#release()})</li>
     *     <li>{@linkplain #PREPARING} if {@linkplain #setupMediaPlayer()} has been called, and
     *     the class started preparing the media player, but
     *     {@linkplain android.media.MediaPlayer.OnPreparedListener#onPrepared(android.media.MediaPlayer)}
     *     was not yet called</li>
     *     <li>{@linkplain #PREPARED} once the media player is prepared and
     *     {@linkplain android.media.MediaPlayer.OnPreparedListener#onPrepared(android.media.MediaPlayer)}
     *      was called.
     *     </li>
     * </ul>
     */
    private enum PlayerStates {
        UNINIT, PREPARING, PREPARED;
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        CastApplication.getPlayerManager().setMediaPlayer(this);

        Log.d(TAG, "New LocalPlayerActivity created");

        loadViews();
        setupActionBar();
        setupControlsCallbacks();

        // see what we need to play and were
        Bundle b = getIntent().getExtras();
        if (b != null) {
            selectedMediaInfo = com.google.android.libraries.cast.companionlibrary.utils.Utils.bundleToMediaInfo(getIntent().getBundleExtra("media"));
            stream = StreamManager.getStreamFromMediaInfo(selectedMediaInfo);
            initialPosition = b.getInt("position");

            Log.d(TAG, "Setting url of the media player to: " + selectedMediaInfo.getContentId());

            // read custom data to configure the playback type and appropriate seek action
            readMediaInfo();

            //LIVE CONTENT
            if (isLive){
                Log.d(TAG, "Live setup.");
                setupMediaPlayer();
            }

            // VOD CONTENT
            else{
                // VOD AS LIVE
                if(seekType == SeekType.OSA_SEEK) {
                    if(initialPosition != 0) {
                        Log.d(TAG, "VOD setup. Initial position: " + initialPosition);
                        CastApplication.getStreamManager().seek(stream, initialPosition);
                    } else {
                        Log.d(TAG, "VOD setup. from start");
                        setupMediaPlayer();
                    }
                }
                else{
                    Log.d(TAG, "Native setup.");
                    if (initialPosition != 0)
                        setupMediaPlayer();
                    else
                        setupMediaPlayer();
                }
            }

//            if (!isLive && seekType == SeekType.OSA_SEEK && initialPosition != 0) {
//                CastApplication.getStreamManager().seek(stream, initialPosition);
//            } else {
//                setupMediaPlayer();
//            }

            setupSeekbarListener();
            playbackState = PlaybackState.PAUSED;
            updatePlayButton(playbackState);
        }
        if (null != titleView) {
            updateMetadata(true);
        }
    }


    /**
     * Updates the mediaplayer and the play button depending on the playback state.
     * Progression: from Paused -> Playing -> Paused ad
     *
     */
    private void togglePlayback() {
        stopControllersTimer();
        switch (playbackState) {
            case PAUSED:
                if (mediaPlayerState == PlayerStates.PREPARED) {
                    mediaPlayer.start();
                    playbackState = PlaybackState.PLAYING;

                    startControllersTimer();
                    restartTrickplayTimer();
                }
                break;

            case PLAYING:
                playbackState = PlaybackState.PAUSED;
                mediaPlayer.pause();
                break;

            case IDLE_FINISHED:
                setupMediaPlayer();

                mediaPlayer.start();
                playbackState = PlaybackState.PLAYING;
                restartTrickplayTimer();

                break;

            default:
                break;
        }
        updatePlayButton(playbackState);
    }

    /**
     * Setup the media player and prepare it for playing.
     * This method is called whenever the media info {@linkplain #selectedMediaInfo} is updated or
     * when the activity is created.
     */
    private void setupMediaPlayer() {
        try {
            if (this.mediaPlayerState == PlayerStates.PREPARING) {
                Log.e(TAG, "Cannot setup media player while media player is preparing! ");
                return;
            }

            Log.d(TAG, "Setting up the media player");

            // set the state as Preparing.
            this.mediaPlayerState = PlayerStates.PREPARING;

            if (mediaPlayer != null) {
                try {
                    mediaPlayer.reset();
                    mediaPlayer.release();
                    mediaPlayer = null;
                } catch (IllegalStateException ile) {
                    Log.e(TAG, "Error: " + ile.getMessage());
                }
            }

            mediaPlayer = new MediaPlayer();
            setupControlsCallbacks();


            HashMap<String, String> headers = new HashMap<String, String>();
            if (selectedMediaInfo.getCustomData() != null && selectedMediaInfo.getCustomData().has(StreamPlayerWrapper.CC_STRING))
                headers.put("Cookie", selectedMediaInfo.getCustomData().optString(StreamPlayerWrapper.CC_STRING));

            mediaPlayer.setDataSource(getApplicationContext(), Uri.parse(selectedMediaInfo.getContentId()), headers);
            mediaPlayer.setAudioStreamType(AudioManager.STREAM_MUSIC);

            mediaPlayer.prepareAsync();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void surfaceCreated(SurfaceHolder holder) {
        Log.d(TAG, "Surface created");
        mediaPlayer.setDisplay(surfaceHolder);
    }

    @Override
    public void surfaceChanged(SurfaceHolder holder, int format, int width, int height) {
    }

    @Override
    public void surfaceDestroyed(SurfaceHolder holder) {
        Log.d(TAG, "Surface destroyed");
    }

    private void stopTrickplayTimer() {
        Log.d(TAG, "Stopped TrickPlay Timer");
        if (null != seekbarTimer) {
            seekbarTimer.cancel();
        }
    }

    private void restartTrickplayTimer() {
        stopTrickplayTimer();
        seekbarTimer = new Timer();
        seekbarTimer.scheduleAtFixedRate(new UpdateSeekbarTask(), 100, 1000);
        Log.d(TAG, "Restarted TrickPlay Timer");
    }

    private void stopControllersTimer() {
        if (null != controllersTimer) {
            controllersTimer.cancel();
        }
    }

    private void startControllersTimer() {
        if (null != controllersTimer) {
            controllersTimer.cancel();
        }

        controllersTimer = new Timer();
        controllersTimer.schedule(new HideControllersTask(), 5000);
    }

    // should be called from the main thread
    private void updateControllersVisibility(boolean show) {
        if (show) {
            getSupportActionBar().show();
            controllersView.setVisibility(View.VISIBLE);
        } else {
            getSupportActionBar().hide();
            controllersView.setVisibility(View.INVISIBLE);
        }
    }

    @Override
    protected void onPause() {
        super.onPause();
        Log.d(TAG, "onPause() was called");
        if (null != seekbarTimer) {
            seekbarTimer.cancel();
            seekbarTimer = null;
        }
        if (null != controllersTimer) {
            controllersTimer.cancel();
        }

        if (mediaPlayer != null && mediaPlayerState == PlayerStates.PREPARED && mediaPlayer.isPlaying()) {
            mediaPlayer.pause();
        }

        playbackState = PlaybackState.PAUSED;
        updatePlayButton(PlaybackState.PAUSED);
        CastApplication.getCastManager().decrementUiCounter();
    }

    @Override
    protected void onStop() {
        Log.d(TAG, "onStop() was called");
        super.onStop();
    }

    @Override
    protected void onDestroy() {
        Log.d(TAG, "onDestroy() is called");

        stopControllersTimer();
        stopTrickplayTimer();

        if (mediaPlayer != null) {
            mediaPlayer.release();
        }

        super.onDestroy();
    }

    @Override
    protected void onStart() {
        Log.d(TAG, "onStart was called");
        super.onStart();
    }

    @Override
    protected void onResume() {
        Log.d(TAG, "onResume() was called");
        CastApplication.getCastManager().incrementUiCounter();
        super.onResume();
    }

    /**
     * Timer task to hide the controller buttons after the media player has started playing
     */
    private class HideControllersTask extends TimerTask {
        @Override
        public void run() {
            handler.post(new Runnable() {
                @Override
                public void run() {
                    updateControllersVisibility(false);
                    areControllersVisible = false;
                }
            });

        }
    }

    private class UpdateSeekbarTask extends TimerTask {

        @Override
        public void run() {
            handler.post(new Runnable() {
                @Override
                public void run() {
                    int currentPos = 0;
                    if (mediaPlayer != null && mediaPlayerState == PlayerStates.PREPARED) {
                        currentPos = mediaPlayer.getCurrentPosition();

                        // in case the seekOffset is not negative, we have to add it to the currentPos.
                        // this is because the seek offset is non-negative when the local player receives a new URL
                        // after the it initiated a seek operation. The seek offset represents the offset of the playing position
                        // in the new file.
                        if (seekOffset != -1) {
                            currentPos += seekOffset;
                        }

                        // in case we play locally the same stream after playing it on ChromeCast,
                        // the initialPosition will be positive, and we use this to properly display
                        // progress in the video
//                        if (initialPosition >= 0 && !isLive && seekType == SeekType.OSA_SEEK) {
//                            currentPos += initialPosition;
//                        }
                    }
                    updateSeekbar(currentPos, duration);
                }
            });
        }
    }


    private void setupControlsCallbacks() {
        mediaPlayer.setOnErrorListener(new OnErrorListener() {

            @Override
            public boolean onError(MediaPlayer mp, int what, int extra) {
                Log.e(TAG, "OnErrorListener.onError(): VideoView encountered an " +
                        "error, what: " + what + ", extra: " + extra);
                String msg = "";
                if (extra == MediaPlayer.MEDIA_ERROR_TIMED_OUT) {
                    msg = getString(R.string.video_error_media_load_timeout);
                } else if (extra == MediaPlayer.MEDIA_ERROR_IO) {
                    msg = getString(R.string.video_error_media_error_io);
                } else if (extra == MediaPlayer.MEDIA_ERROR_MALFORMED) {
                    Log.e(TAG, getString(R.string.video_error_media_error_malformed));
                } else if (extra == MediaPlayer.MEDIA_ERROR_UNSUPPORTED) {
                    Log.e(TAG, getString(R.string.video_error_media_unsupported));
                } else if (what == MediaPlayer.MEDIA_ERROR_SERVER_DIED) {
                    msg = getString(R.string.video_error_server_unaccessible);
                } else {
                    msg = getString(R.string.video_error_unknown_error);
                }
                Utils.showErrorDialog(LocalPlayerActivity.this, msg);
                if (mediaPlayerState == PlayerStates.PREPARED && mediaPlayer.isPlaying()) {
                    mediaPlayer.stop();
                }

                playbackState = PlaybackState.IDLE_UNKNOWN;
                updatePlayButton(playbackState);
                setupMediaPlayer();
                return true;
            }
        });

        // called once the media player is prepared and ready to play
        mediaPlayer.setOnPreparedListener(new OnPreparedListener() {

            @Override
            public void onPrepared(MediaPlayer mp) {
                Log.d(TAG, "onPrepared is reached");
                mediaPlayer.setDisplay(surfaceHolder);

                mediaPlayerState = PlayerStates.PREPARED;
                if (android.os.Build.VERSION.SDK_INT < android.os.Build.VERSION_CODES.JELLY_BEAN_MR1) { 
                    Log.d(MPL_TAG, "Media playback STARTED " + selectedMediaInfo.getContentId());
                    CastApplication.getStreamManager().playbackStarted(stream);
                }

                updateEndText(com.google.android.libraries.cast.companionlibrary.utils.Utils
                        .formatMillis(duration));

                seekBar.setMax(duration);

                playbackState = PlaybackState.PLAYING;
                updatePlayButton(playbackState);

                if (notYetPreparedForFirstTime && initialPosition != 0 && seekType != SeekType.OSA_SEEK) {
                    notYetPreparedForFirstTime = false;
                    seek(initialPosition);
                } else {
                    notYetPreparedForFirstTime = false;
                }

                mediaPlayer.start();

                startControllersTimer();
                restartTrickplayTimer();
            }
        });

        mediaPlayer.setOnVideoSizeChangedListener(new MediaPlayer.OnVideoSizeChangedListener() {
            @Override
            public void onVideoSizeChanged(MediaPlayer mp, int width, int height) {
                Log.d(TAG, "Video size changed!");
                RelativeLayout.LayoutParams vLayout = (RelativeLayout.LayoutParams) surfaceView.getLayoutParams();

                float aspectRatio = (float) width / height;
                float screenRatio = (float) vLayout.width / vLayout.height;

                float topMargin = 0, leftMargin = 0;

                if (screenRatio < aspectRatio)
                    topMargin = (float) vLayout.height - ((float) vLayout.width / aspectRatio);
                else if (screenRatio > aspectRatio)
                    leftMargin = (float) vLayout.width - (vLayout.height * aspectRatio);

                vLayout.setMargins((int) leftMargin, (int) topMargin, 0, 0);
                surfaceView.setLayoutParams(vLayout);
            }
        });

        mediaPlayer.setOnCompletionListener(new OnCompletionListener() {
            @Override
            public void onCompletion(MediaPlayer mp) {
                Log.d(MPL_TAG, "Media playback COMPLETED " + selectedMediaInfo.getContentId());
                CastApplication.getStreamManager().playbackFinished(stream);

                stopTrickplayTimer();
                playbackState = PlaybackState.IDLE_FINISHED;
                updatePlayButton(PlaybackState.IDLE_FINISHED);
            }
        });

        mediaPlayer.setOnSeekCompleteListener(new MediaPlayer.OnSeekCompleteListener() {
            @Override
            public void onSeekComplete(MediaPlayer mp) {
                Log.d(MPL_TAG, "Media player seek complete.");
                // we start only if it a native seek. Otherwise, we wait for the new URL.
                if (seekType == SeekType.NATIVE_SEEK) {
                    mediaPlayer.start();
                }
            }
        });

        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.JELLY_BEAN_MR1)
            mediaPlayer.setOnInfoListener(new MediaPlayer.OnInfoListener() {
                @Override
                public boolean onInfo(MediaPlayer mp, int what, int extra) {
                    if (MediaPlayer.MEDIA_INFO_VIDEO_RENDERING_START == what) {
                        Log.d(MPL_TAG, "Media playback STARTED RENDERING" + selectedMediaInfo.getContentId());
                            CastApplication.getStreamManager().playbackStarted(stream);
                    }
                    return false;
                }
            });

        surfaceView.setOnTouchListener(new OnTouchListener() {

            @Override
            public boolean onTouch(View v, MotionEvent event) {
                if (!areControllersVisible) {
                    updateControllersVisibility(true);
                }
                startControllersTimer();
                return false;
            }
        });

        playPause.setOnClickListener(new OnClickListener() {

            @Override
            public void onClick(View v) {
                togglePlayback();
            }
        });


    }

    /**
     * Sets up the seekbar listeners, depending on the seek type:
     * <ul>
     *     <li>{@linkplain octoshape.android.stream.SeekType#NO_SEEK}: no seeking is allowed</li>
     *     <li>{@linkplain octoshape.android.stream.SeekType#NATIVE_SEEK}: seeking is performed locally, by the {@linkplain #mediaPlayer}. In case the media player was playing, then also restart the timers.</li>
     *     <li>{@linkplain octoshape.android.stream.SeekType#OSA_SEEK}: reset the player, and request to the {@linkplain octoshape.android.stream.StreamManager} to seek the stream to the new position. This will result in issuing a new </li>
     * </ul>
     * </p>
     */
    private void setupSeekbarListener() {
        seekBar.setOnSeekBarChangeListener(new OnSeekBarChangeListener() {
            Integer positionBeforeSeek;

            @Override
            public void onStopTrackingTouch(SeekBar seekBar) {
                Log.d(TAG, "onStopTrackingTouch!");

                Integer finalPosition = seekBar.getProgress();
                int currentLatency = finalPosition - positionBeforeSeek;

                if (currentLatency != 0) {
                    if (seekType == SeekType.NO_SEEK) {
                        Log.d(TAG, "NO SEEKing");
                        startControllersTimer();
                        return;
                    } else {
                        seek(seekBar.getProgress());
                    }

//                if (currentLatency != 0) {
//                    if (seekType == SeekType.NO_SEEK) {
//                        Log.d(TAG, "NO SEEKing");
//                        startControllersTimer();
//                        return;
//                    } else if (seekType == SeekType.NATIVE_SEEK) {
//                        Log.d(TAG, "NATIVE SEEKing: " + seekBar.getProgress());
//                        if (playbackState == PlaybackState.PLAYING) {
//                            startControllersTimer();
//
//                            mediaPlayer.seekTo(seekBar.getProgress());
//                            restartTrickplayTimer();
//                        } else if (playbackState != PlaybackState.IDLE_FINISHED && playbackState != PlaybackState.IDLE_UNKNOWN) {
//                            mediaPlayer.seekTo(seekBar.getProgress());
//                        }
//                        startControllersTimer();
//                    } else if (seekType == seekType.OSA_SEEK) {
//                        Log.d(TAG, "OSA SEEKing: " + seekBar.getProgress());
//                        if (mediaPlayer != null && mediaPlayerState == PlayerStates.PREPARED) {
//                            CastApplication.getStreamManager().playbackSeeking(stream);
//                            mediaPlayer.reset();
//                        }
//
//                        CastApplication.getStreamManager().seek(stream, seekBar.getProgress());
//                    }

                    if (currentLatency < 0) {
                        // rewinding
                        Log.d(TAG, "Rewinding: " + currentLatency);
                    } else {
                        // fast forwarding
                        Log.d(TAG, "Fast Forwarding: " + currentLatency);
                    }

                    latency += currentLatency;
                }

            }

            @Override
            public void onStartTrackingTouch(SeekBar seekBar) {
                positionBeforeSeek = seekBar.getProgress();
                stopTrickplayTimer();
                if (mediaPlayer.isPlaying()) {
                    mediaPlayer.pause();
                }
                stopControllersTimer();
            }

            @Override
            public void onProgressChanged(SeekBar seekBar, int progress,
                                          boolean fromUser) {
                updateStartText(com.google.android.libraries.cast.companionlibrary.utils.Utils
                        .formatMillis(progress));
            }
        });
    }

    private void seek(int time) {
        if (seekType == SeekType.NO_SEEK) {
            Log.d(TAG, "NO SEEKing");
            startControllersTimer();
            return;
        } else if (seekType == SeekType.NATIVE_SEEK) {
            Log.d(TAG, "NATIVE SEEKing: " + time);
            if (playbackState == PlaybackState.PLAYING) {
                startControllersTimer();

                mediaPlayer.seekTo(time);
                restartTrickplayTimer();
            } else if (playbackState != PlaybackState.IDLE_FINISHED && playbackState != PlaybackState.IDLE_UNKNOWN) {
                mediaPlayer.seekTo(time);
            }
            startControllersTimer();
        } else if (seekType == seekType.OSA_SEEK) {
            Log.d(TAG, "OSA SEEKing: " + time);
            if (mediaPlayer != null && mediaPlayerState == PlayerStates.PREPARED) {
                CastApplication.getStreamManager().playbackSeeking(stream);
                mediaPlayer.reset();
            }

            CastApplication.getStreamManager().seek(stream, time);
        }
    }

    /**
     * Updates the seekbar position and seekbar text, depending on the following rules.
     * <ul>
     *     <li>{@linkplain octoshape.android.stream.SeekType#NO_SEEK}: Seekbar is disabled. No seeking allowed.</li>
     *     <li>Otherwise, enable the seekbar.</li>
     *     <li>if <code>isLive == true</code> and there's no latency, set the end text as "Live" and set the progress to the far right end.</li>
     *     <li>if <code>isLive == true</code> and there is latency, set the end text as <code>-mm:ss</code>, where <code>mm:ss</code> is the amount of time the playback is behind the live stream.</li>
     *     <li>if <code>isLive == false</code> set the end text as the duration of the video.</li>
     * </ul>
     *
     * @param position
     * @param duration
     */
    private void updateSeekbar(int position, int duration) {
//        Log.d(TAG, "Seekbar: " + position + " [ " + duration + " ] ");
        if (seekType == SeekType.NO_SEEK || duration == 0) {
            seekBar.setEnabled(false);
        } else {
            seekBar.setEnabled(true);
        }

        seekBar.setMax(duration);

        if (isLive && latency == 0) {
//            Log.d(TAG, "Update seekbar (live & zero latency)");
            seekBar.setProgress(seekBar.getMax());

            updateStartText(com.google.android.libraries.cast.companionlibrary.utils.Utils.formatMillis(position));
            //updateEndText("Live");
            updateEndText("");
        } else if (isLive && latency != 0) {
//            Log.d(TAG, "Update seekbar (live & " + latency + "latency");
            seekBar.setProgress(seekBar.getMax() + latency);

            updateStartText(com.google.android.libraries.cast.companionlibrary.utils.Utils.formatMillis(position));

            String sign = "-";
            if (latency >= 0) {
                sign = "";
            }

            updateEndText(sign + com.google.android.libraries.cast.companionlibrary.utils.Utils.formatMillis(Math.abs(latency)));
        } else {
//            Log.d(TAG, "Updating seekbar (not live)");
            seekBar.setProgress(position);

            updateStartText(com.google.android.libraries.cast.companionlibrary.utils.Utils.formatMillis(position));
            updateEndText(com.google.android.libraries.cast.companionlibrary.utils.Utils.formatMillis(duration));
        }
    }

    private void updateEndText(String text) {
        if (isLive && seekType == SeekType.NO_SEEK) {
            //endText.setText(getString(R.string.stream_live));
            endText.setText("");
        } else {
            endText.setText(text);
        }
    }

    private void updateStartText(String text) {
        if (isLive) {
            startText.setText(getString(R.string.start_text_live_stream));
        } else {
            startText.setText(text);
        }
    }

    private void updatePlayButton(PlaybackState state) {
        switch (state) {
            case PLAYING:
                loadingProgressBar.setVisibility(View.INVISIBLE);
                playPause.setVisibility(View.VISIBLE);
                playPause.setImageDrawable(
                        getResources().getDrawable(R.drawable.ic_av_pause_dark));
                break;
            case PAUSED:

            case IDLE_FINISHED:
            case IDLE_UNKNOWN:
                loadingProgressBar.setVisibility(View.INVISIBLE);
                playPause.setVisibility(View.VISIBLE);
                playPause.setImageDrawable(
                        getResources().getDrawable(R.drawable.ic_av_play_dark));
                break;
            case BUFFERING:
                playPause.setVisibility(View.INVISIBLE);
                loadingProgressBar.setVisibility(View.VISIBLE);
                break;
            default:
                break;
        }
    }

    @SuppressLint("NewApi")
    @Override
    public void onConfigurationChanged(Configuration newConfig) {
        super.onConfigurationChanged(newConfig);
        if (newConfig.orientation == Configuration.ORIENTATION_LANDSCAPE) {
            getWindow().clearFlags(WindowManager.LayoutParams.FLAG_FORCE_NOT_FULLSCREEN);
            getWindow().setFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN,
                    WindowManager.LayoutParams.FLAG_FULLSCREEN);
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.ICE_CREAM_SANDWICH) {
                getWindow().getDecorView().setSystemUiVisibility(View.SYSTEM_UI_FLAG_LOW_PROFILE);
            }
            updateMetadata(false);
            containerView.setBackgroundColor(getResources().getColor(R.color.black));

        } else {
            getWindow().setFlags(WindowManager.LayoutParams.FLAG_FORCE_NOT_FULLSCREEN,
                    WindowManager.LayoutParams.FLAG_FORCE_NOT_FULLSCREEN);
            getWindow().clearFlags(
                    WindowManager.LayoutParams.FLAG_FULLSCREEN);
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.ICE_CREAM_SANDWICH) {
                getWindow().getDecorView().setSystemUiVisibility(View.SYSTEM_UI_FLAG_VISIBLE);
            }
            updateMetadata(true);
            containerView.setBackgroundColor(getResources().getColor(R.color.white));

        }
    }

    private void updateMetadata(boolean visible) {
        if (!visible) {
            descriptionView.setVisibility(View.GONE);
            titleView.setVisibility(View.GONE);
            authorView.setVisibility(View.GONE);
            displaySize = Utils.getDisplaySize(this);
            RelativeLayout.LayoutParams lp = new
                    RelativeLayout.LayoutParams(displaySize.x,
                    displaySize.y + getSupportActionBar().getHeight());
            lp.addRule(RelativeLayout.CENTER_IN_PARENT);
            surfaceView.setLayoutParams(lp);
            surfaceView.invalidate();
        } else {
            MediaMetadata mm = selectedMediaInfo.getMetadata();
            descriptionView.setText(mm.getString(MediaMetadata.KEY_STUDIO));
            titleView.setText(mm.getString(MediaMetadata.KEY_TITLE));
            authorView.setText(mm.getString(MediaMetadata.KEY_SUBTITLE));
            descriptionView.setVisibility(View.VISIBLE);
            titleView.setVisibility(View.VISIBLE);
            authorView.setVisibility(View.VISIBLE);
            displaySize = Utils.getDisplaySize(this);
            RelativeLayout.LayoutParams lp = new
                    RelativeLayout.LayoutParams(displaySize.x,
                    (int) (displaySize.x * aspectRatio));
            lp.addRule(RelativeLayout.ALIGN_PARENT_TOP);
            surfaceView.setLayoutParams(lp);
            surfaceView.invalidate();
        }
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        super.onCreateOptionsMenu(menu);
        getMenuInflater().inflate(R.menu.main, menu);
        CastApplication.getCastManager().addMediaRouterButton(menu, R.id.media_route_menu_item);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
            case R.id.action_settings:
                Intent i = new Intent(LocalPlayerActivity.this, CastPreference.class);
                startActivity(i);
                break;
            case android.R.id.home:
                if (mediaPlayerState == PlayerStates.PREPARED) {
                    NavUtils.navigateUpFromSameTask(this);
                }
                return true;
        }

        return super.onOptionsItemSelected(item);
    }

    private void setupActionBar() {
        Toolbar toolbar = (Toolbar) findViewById(R.id.toolbar);
        toolbar.setTitle("");
        setSupportActionBar(toolbar);

        getSupportActionBar().setDisplayHomeAsUpEnabled(true);
        toolbar.setNavigationOnClickListener(new OnClickListener() {

            @Override
            public void onClick(View v) {
                stopPlay();
            }
        });
    }

    private void loadViews() {
        setContentView(R.layout.player_activity);

        surfaceView = (SurfaceView) findViewById(R.id.surfaceView);
        surfaceHolder = surfaceView.getHolder();
        surfaceHolder.addCallback(this);

        mediaPlayer = new MediaPlayer();

        titleView = (TextView) findViewById(R.id.textView1);
        descriptionView = (TextView) findViewById(R.id.textView2);
        descriptionView.setMovementMethod(new ScrollingMovementMethod());
        authorView = (TextView) findViewById(R.id.textView3);
        startText = (TextView) findViewById(R.id.startText);
        endText = (TextView) findViewById(R.id.endText);
        seekBar = (SeekBar) findViewById(R.id.seekBar1);
        playPause = (ImageView) findViewById(R.id.imageView2);
        loadingProgressBar = (ProgressBar) findViewById(R.id.progressBar1);
        controllersView = findViewById(R.id.controllers);
        containerView = findViewById(R.id.container);
    }

    @Override
    public IMediaPlayer.PlaybackState getPlaybackState() {
        return playbackState;
    }

    @Override
    public int getPlaybackPosition() {
    //       Log.d(TAG, "Playback position: " + mediaPlayer.getCurrentPosition() + " and seekbar: " + seekBar.getProgress());
    //        return mediaPlayer.getCurrentPosition();
        return seekBar.getProgress();
    }

    @Override
    public void stopPlay() {
        CastApplication.getPlayerManager().setMediaPlayer(null);
        if (mediaPlayer != null) {
            try {
                mediaPlayer.stop();
            } catch (IllegalStateException ile) {
                Log.e(TAG, "Error: " + ile.getMessage());
                ile.printStackTrace();
            }
            mediaPlayer.release();
            mediaPlayerState = PlayerStates.UNINIT;
        }
        onPause();
        finish();
    }

    @Override
    public void updatedMediaInfo(MediaInfo mediaInfo) {
        this.selectedMediaInfo = mediaInfo;
        readMediaInfo();
        Log.d(TAG, "Update media info. Player state: " + mediaPlayerState + " and is notYetPreparedForFirstTime=" + notYetPreparedForFirstTime);
        if (mediaPlayerState == PlayerStates.PREPARED || notYetPreparedForFirstTime) {
            new Thread(new Runnable() {
                public void run() {
                    setupMediaPlayer();
                }
            }).start();

        }
    }

    @Override
    public void updateDuration(long duration) {
        this.duration = (int) duration;
    }

    /**
     * Read the custom data passed with the media info. This custom data is used to handle the
     * appropriate playback and seeking.
     */
    private void readMediaInfo() {
        if (selectedMediaInfo.getCustomData() == null) {
            // if no custom data provided, assume conservative values.

            Log.d(TAG, "Null custom data?");
            isLive = true;
            seekType = SeekType.NO_SEEK;
            duration = 0;
            seekOffset = 0;
            return;
        }

        try {
            isLive = selectedMediaInfo.getCustomData().getBoolean(StreamPlayerWrapper.IS_LIVE);
        } catch (JSONException e) {
            e.printStackTrace();
        }

        try {
            seekType = SeekType.getSeekType(selectedMediaInfo.getCustomData().getInt(StreamPlayerWrapper.SEEK_TYPE));
        } catch (JSONException e) {
            e.printStackTrace();
        }

        try {
            duration = (int) selectedMediaInfo.getCustomData().getLong(StreamPlayerWrapper.DURATION);
        } catch (JSONException e) {
            e.printStackTrace();
        }

        try {
            if (!isLive) {
                seekOffset = (int) selectedMediaInfo.getCustomData().getLong(StreamPlayerWrapper.SEEK_OFFSET);
            } else {
                seekOffset = -1;
            }
        } catch (JSONException e) {
            e.printStackTrace();
        }
    }
}
